@@ -1,6 +1,6 @@
 package Net::DNS::Nameserver;
 #
-# $Id: Nameserver.pm 688 2007-12-27 20:23:29Z olaf $
+# $Id: Nameserver.pm 749 2008-12-19 15:20:22Z olaf $
 #
 
 use Net::DNS;
@@ -10,39 +10,31 @@ use IO::Select;
 use Carp qw(cluck);
 
 use strict;
-use vars qw($VERSION
- 	    $has_inet6
- 	    @DEFAULT_ADDR       
- 	    $DEFAULT_PORT
- 	    );
+use vars qw(	$VERSION
+ 		$has_inet6
+ 		);
+
+use constant	FORCE_INET4 => 0;
+
+use constant	DEFAULT_ADDR => 0;
+use constant	DEFAULT_PORT => 53;
 
 use constant	STATE_ACCEPTED => 1;
 use constant	STATE_GOT_LENGTH => 2;
 use constant	STATE_SENDING => 3;
 
-$VERSION = (qw$LastChangedRevision: 688 $)[1];
+$VERSION = (qw$LastChangedRevision: 749 $)[1];
+
+
 
-#@DEFAULT_ADDR is set in the BEGIN block 
-$DEFAULT_PORT=53;
- 	    
- 
- 
 BEGIN {
-    my $force_inet4_only=0;
-    
-    if ($force_inet4_only){
- 	$has_inet6=0;
-    }elsif ( eval {require Socket6;} &&
- 	     # INET6 more recent than 2.01 will not work; sorry.
- 	     eval {require IO::Socket::INET6; IO::Socket::INET6->VERSION("2.00");}) {
- 	import Socket6;
- 	$has_inet6=1;
- 	no  strict 'subs';
- 	@DEFAULT_ADDR= ( 0  );
-    }else{
- 	$has_inet6=0;
- 	@DEFAULT_ADDR= ( INADDR_ANY );
-    }
+	if ( FORCE_INET4 ) {
+		$has_inet6 = 0;
+	} elsif ( eval {require IO::Socket::INET6; IO::Socket::INET6->VERSION("2.01");} ) {
+		$has_inet6 = 1;
+	} else {
+		$has_inet6=0;
+	}
 }
 
 
@@ -55,9 +47,7 @@ BEGIN {
 
 sub new {
 	my ($class, %self) = @_;
-
-
-	if (!$self{"ReplyHandler"} || !ref($self{"ReplyHandler"})) {
+	unless ( ref $self{ReplyHandler} ) {
 		cluck "No reply handler!";
 		return;
 	}
@@ -65,10 +55,11 @@ sub new {
 	# local server addresses must also be accepted by a resolver
 	my @LocalAddr = ref $self{LocalAddr} ? @{$self{LocalAddr}} : ($self{LocalAddr});
 	my $resolver = Net::DNS::Resolver->new;
-	$resolver->nameservers(@DEFAULT_ADDR);
+	$resolver->force_v4(1) unless $has_inet6;
+	$resolver->nameservers(undef);
 	my @localaddresses = $resolver->nameservers(@LocalAddr);
 
-	my $port = $self{LocalPort} || $DEFAULT_PORT;
+	my $port = $self{LocalPort} || DEFAULT_PORT;
 
 	my @sock_tcp;	# All the TCP sockets we will listen to.
 	my @sock_udp;	# All the UDP sockets we will listen to.
@@ -76,16 +67,14 @@ sub new {
 	# while we are here, print incomplete lines as they come along.
 	local $| = 1 if $self{Verbose};
 
-	foreach my $addr (@localaddresses){
-
-		print "Setting up listening sockets for $addr...\n" if $self{Verbose};
-
-		print "Creating TCP socket for $addr - " if $self{Verbose};
+	foreach my $addr ( @localaddresses ? @localaddresses : DEFAULT_ADDR ){
 
  	    #--------------------------------------------------------------------------
  	    # Create the TCP socket.
  	    #--------------------------------------------------------------------------
  		
+ 	    print "\nCreating TCP socket $addr#$port - " if $self{Verbose};
+
 	    my $sock_tcp = inet_new(
  						    LocalAddr => $addr,
  						    LocalPort => $port,
@@ -104,7 +93,7 @@ sub new {
  	    # Create the UDP Socket.
  	    #--------------------------------------------------------------------------
  	    
- 	    print "Creating UDP socket for $addr - " if $self{Verbose};
+ 	    print "Creating UDP socket $addr#$port - " if $self{Verbose};
  	    
  	    my $sock_udp = inet_new(
  						   LocalAddr => $addr,
@@ -121,20 +110,17 @@ sub new {
 
  	}
  	
-  	#--------------------------------------------------------------------------
-  	# Create the Select object.
-  	#--------------------------------------------------------------------------
-  
-  	$self{"select"} = IO::Select->new;
- 
- 	foreach my $sock_tcp  (@sock_tcp){
- 	    $self{"select"}->add($sock_tcp);
- 	}
- 
- 	foreach my $sock_udp  (@sock_udp){
- 	    $self{"select"}->add($sock_udp);
- 	}
-  
+	#--------------------------------------------------------------------------
+	# Create the Select object.
+	#--------------------------------------------------------------------------
+
+ 	my $select = $self{select} = IO::Select->new;
+
+ 	$select->add(@sock_tcp);
+ 	$select->add(@sock_udp);
+
+	return undef unless $select->count;
+
 	#--------------------------------------------------------------------------
 	# Return the object.
 	#--------------------------------------------------------------------------
@@ -160,14 +146,14 @@ sub inet_new {
 #------------------------------------------------------------------------------
 
 sub make_reply {
-	my ($self, $query, $peerhost) = @_;
+	my ($self, $query, $peerhost,$conn) = @_;
 	
 	my $reply = Net::DNS::Packet->new();	# create empty reply packet
 	$reply->header->qr(1);
 
 	my $headermask;
 	
-	if (not $query) {
+	unless ($query) {
 		print "ERROR: invalid packet\n" if $self->{"Verbose"};
 		$reply->header->rcode("FORMERR");
 		return $reply;
@@ -180,14 +166,16 @@ sub make_reply {
 
 
 	# question section returned to caller
-	my @q =  $query->question  ;
+	my @q = $query->question;
 	@q=( Net::DNS::Question->new('', 'ANY', 'ANY') ) unless @q;
 
 	$reply->push("question", @q);
 
-	if ($query->header->opcode eq "QUERY") {
+	if ($query->header->opcode eq "QUERY" ||
+	    $query->header->opcode eq "NS_NOTIFY_OP"   #RFC1996
+	    ) {
 		if ($query->header->qdcount == 1) {
-			my ($qr) = $q[0];
+			my ($qr) = @q;
 			my $qname = $qr->qname;
 			my $qtype = $qr->qtype;
 			my $qclass = $qr->qclass;
@@ -197,9 +185,15 @@ sub make_reply {
 			
 			my ($rcode, $ans, $auth, $add);
 			
-			($rcode, $ans, $auth, $add, $headermask) =
-				&{$self->{"ReplyHandler"}}($qname, $qclass, $qtype, $peerhost, $query);
-			
+			if  ($query->header->opcode eq "QUERY"){
+			  ($rcode, $ans, $auth, $add, $headermask) =
+			      &{$self->{"ReplyHandler"}}($qname, $qclass, $qtype, $peerhost, $query, $conn);
+			}else{
+			  $reply->header->rcode("SERVFAIL") unless 
+			     ( ref $self->{"NotifyHandler"} eq "CODE");
+		  ($rcode, $ans, $auth, $add, $headermask) =
+			      &{$self->{"NotifyHandler"}}($qname, $qclass, $qtype, $peerhost, $query, $conn);
+			}
 			print "$rcode\n" if $self->{"Verbose"};
 			
 			$reply->header->rcode($rcode);
@@ -227,6 +221,9 @@ sub make_reply {
 		$reply->header->aa(1) if $headermask->{'aa'};
 		$reply->header->ra(1) if $headermask->{'ra'};
 		$reply->header->ad(1) if $headermask->{'ad'};
+		if (defined $Net::DNS::opcodesbyname{$headermask->{'opcode'}}){
+			$reply->header->opcode( $headermask->{'opcode'} );
+		}
 	}
 	
 	
@@ -312,7 +309,13 @@ sub tcp_connection {
 			my $qbuf = substr($self->{"_tcp"}{$sock}{"inbuffer"}, 0, $self->{"_tcp"}{$sock}{"querylength"});
 			substr($self->{"_tcp"}{$sock}{"inbuffer"}, 0, $self->{"_tcp"}{$sock}{"querylength"}) = "";
 		  	my $query = Net::DNS::Packet->new(\$qbuf);
-		  	my $reply = $self->make_reply($query, $sock->peerhost);
+			my $conn = {
+				    sockhost => $sock->sockhost(),
+				    sockport => $sock->sockport(),
+				    peerhost => $sock->peerhost(),
+				    peerport => $sock->peerport()
+				   };
+		  	my $reply = $self->make_reply($query, $sock->peerhost, $conn);
 		  	if (not defined $reply) {
 		    		print "I couldn't create a reply for $peer. Closing socket.\n"
 		    			if $self->{"Verbose"};
@@ -344,13 +347,18 @@ sub udp_connection {
 	my $buf = "";
 
  	$sock->recv($buf, &Net::DNS::PACKETSZ);
- 	my ($peerhost,$peerport) = ($sock->peerhost, $sock->peerport);
- 
- 	print "UDP connection from $peerhost:$peerport\n" if $self->{"Verbose"};
+ 	my ($peerhost,$peerport,$sockhost) = ($sock->peerhost, $sock->peerport, $sock->sockhost);
+	
+ 	print "UDP connection from $peerhost:$peerport to $sockhost\n" if $self->{"Verbose"};
 
 	my $query = Net::DNS::Packet->new(\$buf);
-
-	my $reply = $self->make_reply($query, $peerhost) || return;
+	my $conn = {
+		    sockhost => $sock->sockhost,
+		    sockport => $sock->sockport,
+		    peerhost => $sock->peerhost,
+		    peerport => $sock->peerport
+		   };
+	my $reply = $self->make_reply($query, $peerhost, $conn) || return;
 	my $reply_data = $reply->data;
 
 	local $| = 1 if $self->{"Verbose"};
@@ -423,7 +431,7 @@ sub loop_once {
 	  # If we have buffered output, then send as much as the OS will accept
 	  # and wait with the rest
 	  my $len = length $self->{"_tcp"}{$s}{"outbuffer"};
-	  my $charssent = $sock->syswrite($self->{"_tcp"}{$s}{"outbuffer"});
+	  my $charssent = $sock->syswrite($self->{"_tcp"}{$s}{"outbuffer"}) || 0;
 	  print "Sent $charssent of $len octets to ",$self->{"_tcp"}{$s}{"peer"},".\n"
 	      if $self->{"Verbose"};
 	  substr($self->{"_tcp"}{$s}{"outbuffer"}, 0, $charssent) = "";
@@ -520,6 +528,9 @@ Creates a nameserver object.  Attributes are:
   LocalPort		Port on which to listen.  	Defaults to 53.
   ReplyHandler		Reference to reply-handling 
 			subroutine			Required.
+  NotifyHandler         Reference to reply-handling
+                        subroutine for queries with
+                        opdcode NS_NOTIFY (RFC1996)
   Verbose		Print info about received 
 			queries.			Defaults to 0 (off).
 
@@ -533,10 +544,10 @@ IPv6 and IPv4);
 
 
 The ReplyHandler subroutine is passed the query name, query class,
-query type and optionally an argument containing header bit settings
-(see below).  It must return the response code and references to the
-answer, authority, and additional sections of the response.  Common
-response codes are:
+query type and optionally an argument containing the peerhost, the
+incoming query, and the name of the incomming socket (sockethost). It
+must return the response code and references to the answer, authority,
+and additional sections of the response.  Common response codes are:
 
   NOERROR	No error
   FORMERR	Format error
@@ -545,7 +556,7 @@ response codes are:
   NOTIMP	Not implemented
   REFUSED	Query refused
 
-For advanced usage there is an optional argument containing an
+For advanced usage it may also contain a headermaks containing an
 hashref with the settings for the C<aa>, C<ra>, and C<ad> 
 header bits. The argument is of the form 
 C<< { ad => 1, aa => 0, ra => 1 } >>. 
@@ -632,8 +643,12 @@ additional filtering on its basis may be applied.
  use warnings;
  
  sub reply_handler {
-	 my ($qname, $qclass, $qtype, $peerhost) = @_;
+	 my ($qname, $qclass, $qtype, $peerhost,$query,$conn) = @_;
 	 my ($rcode, @ans, @auth, @add);
+
+	 print "Received query from $peerhost to ". $conn->{"sockhost"}. "\n";
+	 $query->print;
+
 	 
 	 if ($qtype eq "A" && $qname eq "foo.example.com" ) {
 		 my ($ttl, $rdata) = (3600, "10.1.2.3");
@@ -646,6 +661,7 @@ additional filtering on its basis may be applied.
   	          $rcode = "NXDOMAIN";
 	 }
 	 
+
 	 # mark the answer as authoritive (by setting the 'aa' flag
 	 return ($rcode, \@ans, \@auth, \@add, { aa => 1 });
  }
